/*
    FEARLUS model0-6-2: LandParcel.m
    Copyright (C) 1999-2002  Macaulay Institute
    Copyright (C) 1999-2002  Macaulay Institute
    Copyright (C) 1999-2002  Macaulay Institute

    This file is part of FEARLUS model0-6-2, an agent-based model of land use
    change.

    FEARLUS model0-6-2 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-6-2 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation of the LandParcel class.

*/

#import "LandParcel.h"
#import "BitString.h"
#import "LandManager.h"
#import "LandUse.h"
#import "Environment.h"
#import "LandParcelHistory.h"
#import "FearlusArguments.h"
#import "Parameter.h"
#import "MiscFunc.h"
#import "Tube.h"
#import "Number.h"
#import "Coordinate2D.h"
#import "Topology.h"
#import "Neighbourhood.h"
#import "StrategyColourKey.h"
#import "Coordinate2DList.h"
#import "SubPopulation.h"

/*

getMatchLinesWidth(raster)

Return the width of the land parcel bitstring match lines.

*/

static unsigned getMatchLinesWidth(id r) {
  if([r conformsTo: @protocol(ZoomRaster)]) {
    return [(id <ZoomRaster>)r getZoomFactor] / 4;
  }
  else {
    return 1;
  }
}


@implementation LandParcel

/*

create

Override the superclass create method to implement a unique identifier system.
This uses a static local variable in the create method. Static variables in
methods are shared across all instances of a class.

*/

+create: (id)z {
  LandParcel *new;
  static unsigned counter = 1;

  new = [super create: z];
  new->pin = counter;
  new->nLUChange = 0;		// These should really go in an initialise
  new->nLMgrChange = 0;		// method.
  new->newLandManager = YES;
  new->displayMgrNbr = NO;
  new->displayPhysNbr = NO;
  new->lpBitToDisplay = 0;
  new->landUseMemory = [Tube create: z setLength: 1];
  new->yieldMemory = [Tube create: z setLength: 1];
  counter++;
  return(new);
}

/*

getPIN

Return the unique identification number of this instance.

*/

-(unsigned)getPIN {
  return(pin);
}

/*

setParameters:

Pass in the set of model parameters

*/

-(void)setParameters: (Parameter *)p {
  parameter = p;
  [landUseMemory resetLength: [parameter maximumMemorySize]];
  [yieldMemory resetLength: [parameter maximumMemorySize]];
}

/*

setNextLandUseTypeTo

Set the next land use to that passed in the argument. The next land
use will become the current land use when the updateLandUse method is
called.

*/

-(void)setNextLandUseTypeTo: (LandUse *)lu {
  nextLandUse = lu;
  if(landUse == lu && [FearlusArguments debugParcels]) {
    id historyIndex;
    LandParcelHistory *lph1, *lph2;

    printf("P: Land parcel %u at (%d, %d) yield: %u", pin, x, y, yield);
    historyIndex = [landParcelHistory begin: scratchZone];
    [historyIndex next];
    if([historyIndex getLoc] == Member) {
      lph1 = [historyIndex get];
      printf(" [%u", [lph1 getYield]);
      [historyIndex next];
      if([historyIndex getLoc] == Member) {
	lph2 = [historyIndex get];
	printf(" [%u]", [lph2 getYield]);
      }
      printf("]");
    }
    [historyIndex drop];
    printf(" -- identical nextLandUse: %u\n", [lu getPIN]);
  }
}

/*

setNextLandManagerTo

Set the next land manager of this parcel to that passed in the argument.

*/

-(void)setNextLandManagerTo: (LandManager *)lm {
  nextLandManager = lm;
}

/*

getLandManager

Return a pointer to the current land manager of this land parcel.

*/

-(LandManager *)getLandManager {
  return(landManager);
}

/*

setLandManager

Set the land manager of the land parcel to the argument. Used by land allocator
during initial set up. Subsequent settings of this variable should be done
using the setNextLandManager method

*/

-(void)setLandManager: (LandManager *)lm {
  landManager = lm;
  nextLandManager = lm;		// Also set the next land manager in case it
                                // doesn't change (which, without this line,
				// would result in a null land manager in the
				// update method).
}

/*

getLandUse

Return a pointer to the current land use of this land parcel.

*/

-(LandUse *)getLandUse {
  return(landUse);
}

/*

recallLandUseAgo:

Return the land use of this land parcel from yrs years ago.

*/

-(LandUse *)recallLandUseAgo: (unsigned)yrs {
  return([landUseMemory getObject: yrs]);
}

/*

recallYieldAgo:

Return the yield of this land parcel from yrs years ago.

*/

-(unsigned)recallYieldAgo: (unsigned)yrs {
  return([[yieldMemory getObject: yrs] getUnsigned]);
}

/*

toggleBits

Toggle those bits in the attribute bit string that are set to 1 in the bit
string passed in the argument. This is used by the environment as part of the
clumping operation. The toggle operation is functionally equivalent to XOR:

                |   Original Bit String
                |
                |      0         1
                |
----------------+----------------------
                |
        No = 0  |      0         1
                |
Toggle?         |
                |
       Yes = 1  |      1         0
                |

*/

-(void)toggleBits: (BitString *)bs {
  BitString *temp;

  if([attributes getBitCount] != [bs getBitCount]) {
    fprintf(stderr,
	    "LandParcel.toggleBits called with BitString of incorrect length"
	    "\n");
    fprintf(stderr,
	    "My bits: %u Called with bits: %u\n", [attributes getBitCount],
	    [bs getBitCount]);
    return;
  }
  temp = [BitString create: scratchZone setBitCount: [attributes getBitCount]];
  [temp setToCopyOf: attributes];
  [attributes setToXOROf: temp and: bs];
  [temp drop];
}

/*

setBitLength

This is designed to be an initiation method, which is called just after
creation. It creates a bit string to store the attributes in, setting its
length to the value stored in the argument. It also creates the history lists.

*/

-(void)setBitLength: (unsigned)bl {
  attributes = [BitString create: [self getZone] setBitCount: bl];
  landParcelHistory = [List create: [self getZone]];
}

/*

randomiseBits

Set the bits in the attribute string to random values with an equal probability
for 1 and 0. This is an initiation method.

*/

-(void)randomiseBits {
  [attributes setEachBitTo: 1 withProbability: 0.5];
}

/*

printHistoryToFile

A method to print the history of the land parcel to a file, in a format
readable by a spreadsheet package.

*/

-(void)printHistoryToFile: (FILE *)fp {
  id indx;
  LandParcelHistory *hd;

  for(indx = [landParcelHistory begin: scratchZone], [indx next];
      [indx getLoc] == Member;
      [indx next]) {
    fprintf(fp, "Parcel\t%u\t", pin);
    hd = [indx get];
    [hd printToFile: fp];
    fprintf(fp, "\t%d\t%d\t'", x, y);
    [attributes printToFile: fp];
    fprintf(fp, "%s", [FearlusArguments nl]);
  }
  [indx drop];
}

/*

updateLandUse

Once the Land Managers have decided what the Land Use is to be for the
year, it must be updated. Land Managers set the nextLandUse instance
variable to the Land Use they wish to use for the year by calling the
setNextLandUseTypeTo: method. All Land Managers make the decision
about Land Use on the basis of the same year's data.

*/

-(void)updateLandUse {
  if([FearlusArguments debugParcels]) {
    printf("P: Updated land use of parcel %u at (%d, %d) from %u to %u\n",
	   pin, x, y, [landUse getPIN], [nextLandUse getPIN]);
  }
  if(landUse != nextLandUse) nLUChange++;
  landUse = nextLandUse;
  [landUseMemory pushItem: landUse];
  [landUse incNLandParcels];
}

/*

updateLandManager

Make the next land manager current

*/

-(void)updateLandManager {
  if([FearlusArguments debugParcels]) {
    printf("P: Updated land manager of parcel %u at (%d, %d) from %u to %u\n",
	   pin, x, y, [landManager getPIN], [nextLandManager getPIN]);
  }
  if(landManager != nextLandManager) {
    nLMgrChange++;
    newLandManager = YES;
  }
  else {
    newLandManager = NO;
  }
  landManager = nextLandManager;
}

/*

recordHistory

Record the history of the past year

*/

-(void)recordHistory {
  LandParcelHistory *newHd;

  newHd = [LandParcelHistory create: [self getZone]];
  [newHd initialiseWithYear: [environment getYear]
	 landUse: landUse
	 landManager: landManager
	 yield: yield
	 strategy: [strategy class]];
  [landParcelHistory addFirst: newHd];
  // Stick the history on the FRONT of the list so getLastYield and yieldDown
  // methods look at the past two years, not the first two years!!!!
  if([parameter dumpHistory]
     && [parameter maxHistoryLength] < [landParcelHistory getCount]) {
    [[landParcelHistory removeLast] drop];
  }
}

/*

setLandUseTypeTo

This method is used by the land managers to set the initial land use only.

*/

-(void)setLandUseTypeTo: (LandUse *)lu {
  landUse = lu;
  nextLandUse = lu;		// Also set the next land use in case it
				// doesn't change this time.
  [landUseMemory pushItem: lu];
  [lu incNLandParcels];
}

/*

calculateYield

Compute the yield of the land parcel

*/

-(void)calculateYield {
  Number *n;

  yield = [landUse getMatchWithLandParcel: self
		   climate: [environment getClimate]
		   economy: [environment getEconomy]];
  n = [Number create: [self getZone]];
  [n setUnsigned: yield];
  [yieldMemory pushItemDrop: n];
  if([FearlusArguments debugParcels]) {
    printf("P: New yield for Land Parcel %d at (%d, %d): %u\n",
	   pin, x, y, yield);
  }
}

/*

setXPos

Set the X and Y co-ordinates of this land parcel.

*/

-(void)setXPos: (int)xx YPos: (int)yy {
  x = xx;
  y = yy;
}

/*

setEnvironment

Set the environment the land parcel belongs to.

*/

-(void)setEnvironment: (Environment *)env {
  environment = env;
}

/*

getEnvironment

Return the environment

*/

-(Environment *)getEnvironment {
  return environment;
}

/*

setStrategy:

Set the strategy used to determine the land use (should this be recorded in the
history? probably.)

*/

-(void)setStrategy: (id <Strategy>)s {
  strategy = s;
}

/*

getStrategy

Return the most recent strategy used to determine the land use. (Whether this
applies to the current land use depends on where we are in the schedule.)

*/

-(id <Strategy>)getStrategy {
  return strategy;
}

/*

setMaxGraphColour

Set the maximum colour that can be used for graphing. This is used for creating
grey level graphs over land parcel space of things like average number of land
use changes. This is a naughty observer swarm thing implemented here. The
maximum colour available for graphing is passed down from the ObserverSwarm.

*/

-(void)setMaxGraphColour: (int)col {
  maxGraphColour = col;
}

/*

getAttributes

Return the attribute bit string

*/

-(BitString *)getAttributes {
  return(attributes);
}

/*

getOneAttribute

Return the value of the nth bit in the attribute bit string

*/

-(int)getOneAttribute: (unsigned)a {
  return([attributes getBit: a]);
}

/*

setAttribute:toValue:

Set an attribute in the bit string to a value

*/

-(void)setAttribute: (unsigned)a toValue: (int)v {
  if(v == 1 || v == 0) {
    [attributes setBit: a to: v];
  }
  else {
    fprintf(stderr, "Must use 1 or 0 for value in "
	    "LandParcel.setAttribute:toValue: (%d)\n", v);
    exit(1);
  }
}

/*

getX

Return the X co-ordinate of this land parcel

*/

-(int)getX {
  return(x);
}

/*

getY

Return the Y co-ordinate of this land parcel

*/

-(int)getY {
  return(y);
}

/*

getNextNeighbour

Return the next Neighbour of this land parcel. Uses the environment's get next
neighbour method. This method will only return land parcels. Does a paranoid
check that the neighbour returned by the environment is a land parcel. If not,
it will return false. This is designed to be used in a loop like:

while(aLandParcelPointer = [aLandParcel getNextNeighbour]) {
  do some things to aLandParcelPointer
}

*/

-(LandParcel *)getNextNeighbour {
  id lp;

  lp = [environment getNextNeighbourOfX: x Y: y];
  if(!lp || ([lp class] != [LandParcel class])) {
    return(nil);
  }
  else {
    return(lp);
  }
}

/*

getYield

Return the yield for the current year.

*/

-(unsigned)getYield {
  return(yield);
}

/*

getIntYield

Return the yield for the current year as an integer (maybe needed for graphs)

*/

-(int)getIntYield {
  return((int)yield);
}

/*

getLastYield

Get the yield for last year. If there isn't a last year, return -1 (which is,
of course, 4294967295 -- an enormous bit string for the land use is needed
before yields of this magnitude are likely).

*/

-(unsigned)getLastYield {
  id historyIndex;
  LandParcelHistory *lastYear;

  historyIndex = [landParcelHistory begin: scratchZone];
  lastYear = [historyIndex next];
  if([historyIndex getLoc] == Member) {
    [historyIndex drop];
    return([lastYear getYield]);
  }
  else {
    [historyIndex drop];
    return(-1);
  }
}

/*

yieldDown

Return 1 if the yield is going down (last year's yield is < year before last's,
0 if not, and -1 if we haven't got two years of yield history yet.

*/

-(int)yieldDown {
  id historyIndex;
  LandParcelHistory *lastYear, *yearBeforeLast;

  historyIndex = [landParcelHistory begin: scratchZone];
  lastYear = [historyIndex next];
  if([historyIndex getLoc] != Member) {
    [historyIndex drop];
    return(-1);
  }
  yearBeforeLast = [historyIndex next];
  if([historyIndex getLoc] != Member) {
    [historyIndex drop];
    return(-1);
  }
  [historyIndex drop];
  if([lastYear getYield] < [yearBeforeLast getYield]) return(1);
  else return(0);
}

/*

setChangeLURank:

Set the rank of this land parcel in terms of number of changes of land use. The
rank should be normalised by the number of land parcels (and so will be a
number between 0 (highest rank) and 1 (lowest rank)

*/

-(void)setChangeLURank: (double)rank {
  changeLURank = rank;
}

/*

setChangeLMgrRank:

Set the rank of this land parcel in terms of the number of changes of land
manager. Rank should be a number between 0 and 1.

*/

-(void)setChangeLMgrRank: (double)rank {
  changeLMgrRank = rank;
}

/*

getNLUChange

Return the number of land use changes for this land parcel

*/

-(unsigned)getNLUChange {
  return(nLUChange);
}

/*

getNLMgrChange

Return the number of changes of land manager for this land parcel

*/

-(unsigned)getNLMgrChange {
  return(nLMgrChange);
}

/*

hasANewLandManager

Return a boolean value indicating whether or not the current land manager is
a different land manager from last year.

*/

-(BOOL)hasANewLandManager {
  return(newLandManager);
}

/*

determineNWNeighbourAttributeMatch;

Determine the degree of match between the bit string of this land parcel and
the bit string of the neighbouring land parcels to the north and west. This can
then be used to draw lines between the land parcels giving a rough indication
of how similar their bit strings are. We only need to do north and west,
because the south will be taken care of by the land parcel below and the east
by the land parcel to the right.

*/

-(void)determineNWNeighbourAttributeMatch {
  northNeighbourMatch = [[self getAttributes]
			  getMatchWith:
			    [[environment getObjectAtX: x
					 Y: [MiscFunc
					      Mod: (y - 1)
					      inRange0To:
						[environment getSizeY]]]
			      getAttributes]];
  westNeighbourMatch = [[self getAttributes]
			 getMatchWith:
			   [[environment
			      getObjectAtX: [MiscFunc
					     Mod: (x - 1)
					     inRange0To:
					       [environment getSizeX]]
			      Y: y]
			      getAttributes]];
}

/*

toggleMgrNbrDisplay

Toggle the flag used to indicate whether or not the social neighbourhood of
the land manager who owns this parcel is to be displayed.

*/

-(void)toggleMgrNbrDisplay {
  displayMgrNbr = displayMgrNbr ? NO : YES;
}

/*

togglePhysNbrDisplay

Toggle the flag used to indicate whether or not the physical neighbourhood of
this land parcel is to be displayed.

*/

-(void)togglePhysNbrDisplay {
  displayPhysNbr = displayPhysNbr ? NO : YES;
}

/*

drawPhysNbrsOf:on:

Draw the physical neighbours of a given land parcel on the specified
raster.  There are complications here because the co-ordinate to pass
to the neighbourhood function is not necessarily that of self. The
topology passes negative co-ordinates and co-ordinates larger than the
size of the environment if it is toroidal. The neighbourhood function
does not know about toroidal or bounded environments -- it just
implements simple rules about whether or not two co-ordinates (as
integers) are neighbours given that particular function. So, first we
need to go to the topology object and ask for our co-ordinate relative
to the land parcel -- this will also tell us if we're in range. Then
we can call the neighbourhood function to ask if each of our von
Neumann neighbours are neighbours of the land parcel. If they are
neighbours, then call them and ask them to draw their neighbours (but
only if we are nearer to the original land parcel than the prospective
neighbour). If not, then draw a line between us and the von Neumann
neighbour.

*/

-drawPhysNbrsOf: (LandParcel *)lp
	     on: (id <Raster>)aRaster
	 memory: (Coordinate2DList *)clist {
  Coordinate2D *coord, *nbrcoord;

  if([FearlusArguments debugInnerCoreMethods]) {
    printf("LandParcel at (%d, %d) drawing nbrs of (%d, %d)\n", x, y,
	   [lp getX], [lp getY]);
  }
  coord = [Coordinate2D create: scratchZone setX: 0 Y: 0];
  nbrcoord = [Coordinate2D create: scratchZone setX: 0 Y: 0];
  if([[environment getTopology] getNeighbourhoodCoordinatesOf: self
				relativeTo: lp in: coord]) {
    int xx, yy, c;
    id <Neighbourhood> nbrFunc = [environment getNeighbourhoodFunction];

    if([clist member: coord]) {
      if([FearlusArguments debugInnerCoreMethods]) {
	printf("Already visited (%d, %d):\n", x, y);
      }
      return self;
    }
    [clist addHead: coord];
    if([FearlusArguments debugInnerCoreMethods]) {
      printf("Converted my co-ordinates to: (%d, %d)\n", coord->x, coord->y);
    }
    /* This loop loops through all the von Neumann relative co-ordinates like
       this:

       c = 0:	xx = -1, yy =  0
       c = 1:   xx =  0, yy = ((1%2)-(2*(1/2)))*(1%2) = (1-0)*1 =  1
       c = 2:	xx =  1, yy = ((2%2)-(2*(2/2)))*(2%2) = (0-2)*0 =  0
       c = 3:	xx =  0, yy = ((3%2)-(2*(3/2)))*(3%2) = (1-2)*1 = -1

    */
    for(xx = -1, yy = 0, c = 0;
	c < 4;
	c++, xx = yy, yy = ((c % 2) - (2 * (c / 2))) * (c % 2)) {
      nbrcoord->x = coord->x + xx;
      nbrcoord->y = coord->y + yy;
      if([nbrFunc areNeighbours: lp and: nbrcoord]) {
	if([FearlusArguments debugInnerCoreMethods]) {
	  printf("(%d, %d) The neighbouring cell at (%d, %d) is a neighbour of"
		 " (%d, %d)\n", x, y, nbrcoord->x, nbrcoord->y, [lp getX],
		 [lp getY]);
	}
	[[environment getLandParcelAtX: nbrcoord->x Y: nbrcoord->y]
	  drawPhysNbrsOf: lp on: aRaster memory: clist];
      }
      else {
	/* Draw a line between me and my neighbour. The little formulae for
	   the co-ordinates of the line can be explained by referring to the
	   diagram below (remembering that x,y refers to the top left hand
	   corner of the cell, and the origin is in the top left hand corner):

                      xx=0,yy=-1 line: (x,y) to: (x+1,y+1)
                                      ^
				      |
 xx=-1,yy=0 line: (x,y) to: (x,y+1) <-+-> xx=1,yy=0 line: (x+1,y) to (x+1,y+1)
	                              |
				      v
                       xx=0,yy=1 line: (x,y+1) to: (x+1,y+1)
	*/
	if([FearlusArguments debugInnerCoreMethods]) {
	  printf("(%d, %d) The neighbouring cell at (%d, %d) is not a "
		 "neighbour of (%d, %d).\n"
		 "Drawing a line from (%d, %d) to (%d, %d)\n", x, y,
		 nbrcoord->x, nbrcoord->y, [lp getX], [lp getY],
		 (xx < 1 ? x : x + 1), (yy < 1 ? y : y + 1), 
		 (xx < 0 ? x : x + 1), (yy < 0 ? y : y + 1));
	}
	[aRaster lineX0: (xx < 1 ? x : x + 1)
		 Y0: (yy < 1 ? y : y + 1)
		 X1: (xx < 0 ? x : x + 1)
		 Y1: (yy < 0 ? y : y + 1)
		 Width: 2 Color: 0];
      }
    }

  }
  [nbrcoord drop];
  return self;
}

/*

drawSelfOn

Draw the land use of the land parcel.

*/

-drawSelfOn: (id <Raster>)aRaster {
  [aRaster drawPointX: x Y: y Color: [landUse getColour]];
  return([self drawMgrBoundariesOn: aRaster]);
}

/*

drawMatchLinesOn:

Draw lines representing the degree of match between my bitstring and those of
my neighbours at N and W, as calculated in determineNWNeighbourAttributeMatch.
This method relies on the fact that the ZoomRaster object draws a rectangle
from (x,y) to (x+1,y+1) when it does a drawPointX:Y:Color:

*/

-drawMatchLinesOn: (id <Raster>)aRaster {
  unsigned width = getMatchLinesWidth(aRaster);
  unsigned pbs = [parameter parcelBitStringSize];
  
  if(pbs == 0) pbs++;
  [aRaster lineX0: x Y0: y X1: x Y1: y + 1 Width: (int)width
	   Color: (Color)(1 + ((38 * westNeighbourMatch) / pbs))];
  // The naughty "40" constant comes from the number of grey-level colours
  // the observerSwarm attempts to allocate.
  [aRaster lineX0: x Y0: y X1: x + 1 Y1: y Width: (int)width
	   Color: (Color)(1 + ((38 * northNeighbourMatch) / pbs))];
  return self;
}

/*

drawNbrLinesOn:

Draw lines showing the boundary of the physical neighbourhood of this land
parcel if it has been clicked on to do so.

*/

-drawNbrLinesOn: (id <Raster>)aRaster {
  if(displayPhysNbr) {
    Coordinate2DList *clist = [Coordinate2DList create: scratchZone];

    [self drawPhysNbrsOf: self on: aRaster memory: clist];
    [clist deleteAll];
    [clist drop];
    [aRaster lineX0: x Y0: y X1: x + 1 Y1: y + 1 Width: 2 Color: 20];
    [aRaster lineX0: x + 1 Y0: y X1: x Y1: y + 1 Width: 2 Color: 20];
  }
  return self;
}

/*

drawMgrOn

Draw the current owning land manager on a raster and boundary lines

*/

-drawMgrOn: (id <Raster>)aRaster {
  LandManager *northLM, *southLM, *eastLM, *westLM;
  int Nsame, Ssame, Esame, Wsame;

  northLM = [[environment getLandParcelAtX: x Y: y - 1] getLandManager];
  southLM = [[environment getLandParcelAtX: x Y: y + 1] getLandManager];
  eastLM = [[environment getLandParcelAtX: x + 1 Y: y] getLandManager];
  westLM = [[environment getLandParcelAtX: x - 1 Y: y] getLandManager];
  Nsame = (northLM == landManager) ? 1 : 0;
  Ssame = (southLM == landManager) ? 1 : 0;
  Esame = (eastLM == landManager) ? 1 : 0;
  Wsame = (westLM == landManager) ? 1 : 0;

  /* OK now draw the thing. For the boundary lines we depend on the Swarm
     library treating (x, y) as the top left of the cell */

  [aRaster drawPointX: x Y: y Color: [landManager getColour]];

  /* West line */
  [aRaster lineX0: x Y0: y X1: x Y1: y + 1 Width: (Wsame ? 1 : 2)
	   Color: (Wsame ? 38 : 2)];

  /* South line */
  [aRaster lineX0: x Y0: y + 1 X1: x + 1 Y1: y + 1 Width: (Ssame ? 1 : 2)
	   Color: (Ssame ? 38 : 2)];

  /* East line */
  [aRaster lineX0: x + 1 Y0: y + 1 X1: x + 1 Y1: y Width: (Esame ? 1 : 2)
	   Color: (Esame ? 38 : 2)];

  /* North line */
  [aRaster lineX0: x + 1 Y0: y X1: x Y1: y Width: (Nsame ? 1 : 2)
	   Color: (Nsame ? 38 : 2)];

  return(self);
}

/*

drawSubPopOn

Draw the subpopulation on a raster

*/

-drawSubPopOn: (id <Raster>)aRaster {
  [aRaster drawPointX: x Y: y 
	   Color: [[landManager getSubPopulation] getColour]];

  return([self drawMgrBoundariesOn: aRaster]);
}

/*

drawMgrBoundariesOn:

Draw a black boundary between (von Neumann) neighbouring parcels if they have a
different land manager. (Assuming the origin is in the top left of the display
and the co-ordinates of the cell refer to the top left of its square.)

*/

-drawMgrBoundariesOn: (id <Raster>)aRaster {
  LandManager *northLM, *southLM, *eastLM, *westLM;

  northLM = [[environment getLandParcelAtX: x Y: y - 1] getLandManager];
  southLM = [[environment getLandParcelAtX: x Y: y + 1] getLandManager];
  eastLM = [[environment getLandParcelAtX: x + 1 Y: y] getLandManager];
  westLM = [[environment getLandParcelAtX: x - 1 Y: y] getLandManager];
  if(northLM != landManager) {
    [aRaster lineX0: x + 1 Y0: y X1: x Y1: y Width: 1 Color: 1];
  }
  if(southLM != landManager) {
    [aRaster lineX0: x Y0: y + 1 X1: x + 1 Y1: y + 1 Width: 1 Color: 1];
  }
  if(eastLM != landManager) {
    [aRaster lineX0: x + 1 Y0: y + 1 X1: x + 1 Y1: y Width: 1 Color: 1];
  }
  if(westLM != landManager) {
    [aRaster lineX0: x Y0: y X1: x Y1: y + 1 Width: 1 Color: 1];
  }
  return(self);
}

/*

drawNLUChangeOn

Draw the average number of land use changes on a raster -- shown as a grey
scale between 0 and maxGraphColour.

*/

-drawNLUChangeOn: (id <Raster>)aRaster {
  int colour;
  double frequencyOfChange;

  frequencyOfChange = (double)nLUChange / (double)[environment getYear];
				// Assumes year starts at 0. Not necessarily a
				// good idea.
  colour = 1 + (int)(frequencyOfChange * (double)(maxGraphColour - 2));
  [aRaster drawPointX: x Y: y Color: colour];
  return(self);
}

/*

drawNMgrChangeOn

Draw the average number of land manager changes on a raster -- shown as a grey
scale between 0 and maxGraphColour.

*/

-drawNMgrChangeOn: (id <Raster>)aRaster {
  int colour;
  double frequencyOfChange;

  frequencyOfChange = (double)nLMgrChange / (double)[environment getYear];
				// Assumes year starts at 0. Not necessarily a
				// good idea.
  colour = 1 + (int)(frequencyOfChange * (double)(maxGraphColour - 2));
  [aRaster drawPointX: x Y: y Color: colour];
  return(self);
}

/*

drawChangeLURankOn:

Draw the change in land use rank of this land parcel on the raster -- shown as
a grey level between 0 (lowest rank) and maxGraphColour (highest rank).

*/

-drawChangeLURankOn: (id <Raster>)aRaster {
  int colour;

  colour = 1 + (int)((1.0 - changeLURank) * (double)(maxGraphColour - 2));
  [aRaster drawPointX: x Y: y Color: colour];
  return(self);
}

/*

drawChangeLMgrRankOn:

Draw the change in land manager rank of this land parcel on the raster -- shown
as a grey level between 0 (lowest rank) and maxGraphColour (highest rank).

*/

-drawChangeLMgrRankOn: (id <Raster>)aRaster {
  int colour;

  colour = 1 + (int)((1.0 - changeLMgrRank) * (double)(maxGraphColour - 2));
  [aRaster drawPointX: x Y: y Color: colour];
  return(self);
}

/*

drawStrategyOn:

Draw the strategy on the raster passed in as argument

*/

-drawStrategyOn: (id <Raster>)aRaster {
  [aRaster drawPointX: x Y: y
	   Color: [[parameter strategyColourKey]
		    getColourForStrategyClass: [strategy class]]];
  return self;
}

/*

drawBitStringOn:

Draw the value of the bit string on the raster.

*/

-drawBitStringOn: (id <Raster>)aRaster {
  if([attributes getBitCount] == 0) {
    [aRaster drawPointX: x Y: y Color: 0];
  }
  else if([attributes getBit: lpBitToDisplay]) {
    [aRaster drawPointX: x Y: y Color: 1];
  }
  else {
    [aRaster drawPointX: x Y: y Color: 39];
  }
  return self;
}

/*

setBitToDisplay:

*/

-setBitToDisplay: (unsigned)bit {
  lpBitToDisplay = (int)bit % (int)[attributes getBitCount];
  return self;
}

/*

incBitToDisplay

*/

-incBitToDisplay {
  lpBitToDisplay = [MiscFunc Mod: (lpBitToDisplay + 1)
			     inRange0To: (int)[attributes getBitCount]];
  return self;
}

/*

decBitToDisplay

*/

-decBitToDisplay {
  lpBitToDisplay = [MiscFunc Mod: (lpBitToDisplay - 1)
			     inRange0To: (int)[attributes getBitCount]];
  return self;
}

/*

drop

Override the superclass method to make sure all items created by instances of
this class are also dropped.

*/

-(void)drop {
  [landParcelHistory deleteAll];
  [landParcelHistory drop];
  [attributes drop];
  [yieldMemory resetLengthDrop: 0];
  [yieldMemory drop];
  [landUseMemory drop];
  [super drop];
}

@end
